//+------------------------------------------------------------------+
//|                                              NewsTrackerV1_0.mqh |
//|                                               Paul Hampton-Smith |
//|                                               paul1000@pobox.com |
//+------------------------------------------------------------------+
/*


Copy this include file to the experts\include folder. No need to compile

This include file is intended for the following two uses:

EITHER
~~~~~~
to be used with systems that work on the shorter timeframes (1M, 5M and 15M) to 
give them the opportunity to be out of the market during news events.

example:
   #include <NewsTrackerV1_0.mqh>

   string strNews; 
   datetime NewsTime;

   // avoid buying if within 5 minutes of news event
   if (MyBuySignal && !NewsEvent(5,strNews,NewsTime)) 
      DoMyBuying();

   // close all positions if within 5 minutes of news event
   if (OpenOrders() && NewsEvent(5,strNews,NewsTime)) 
      CloseOrDeleteAllOrders();

 OR
~~~~
to be used with systems that work on the news event itself
   
   #include <NewsTrackerV1_0.mqh>

   string strNews; 
   datetime NewsTime;
	double Parameters[MAX_PARAMETERS];
	
   if (NextNews(strNews, NewsTime, Parameters))
   {
      // create a straddle stop order if news will happen in 60 seconds
      if ( CurTime() >= NewsTime-60 && OpenStopOrders == 0 && OpenOrders == 0 )
      {
         CreateStraddle(Parameters);
      }
   }
   

THE FOLLOWING TWO ITEMS NEED TO BE SET UP
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

1. Global variable UTCtoServerTimeInMinutes is added to each instance of MT4 (see menu item Files->Global Variables)

It enables different instances of MT4 using different brokers to run EAs with identical time triggers

Examples:
   UTCtoServerTimeInMinutes = 0   for InterbankFX because their server time = UTC/GMT
   UTCtoServerTimeInMinutes = 180 for FXDD        because their server time = UTC/GMT plus 3 hours


2. File NewsItems.csv is added to the experts\files folder

Format of the file is as per the following example (the header must be included)

Description,Currency,DateTime,TimeZone,Parameter1,Parameter2,Parameter3
News description 1,USD,2006.08.09 12:00,UTC,0.5,0,3
News description 2,GBP,2006.08.09 17:30,UTC,0,100,3
News occurring every day in US,USD,08:30,EST,0,50,4
News occurring every day in GB,GBP,09:30,BST,1,0,3

This format can be read and saved easily as a .csv file in excel

Valid values for each of the fields are

1. Description: string without commas

2. Currency: any 3 alpha currency string

3. DateTime: yyyy.mm.dd hh:mm  for a specific date 
     or      hh:mm  for a time that occurs every Day
     this second option can be a convenient way of ensuring that frequently used news times are always defined 
     without having to update the file. It will therefore work with backtesting.

4. TimeZone
   UTC: Time is in Greenwich Mean Time
   GMT: Time is in Greenwich Mean Time
   BST: Time is in British Summer Time (UTC + 1 hour - London with daylight saving)
   EST: Time is in Eastern Standard Time (UTC - 5 hours - New York)
   EDT: Time is in Eastern Daylight Time (UTC - 4 hours - New York with daylight saving)
   CST: Time is in South Australian CST Time (UTC + 9.5 hours)
   CSST: Time is in South Australian Central Summer Time (UTC + 10.5 hours)

5. Parameters
	Any number of parameters up to MAX_PARAMETERS

The news items can be cached to increase speed for backtesting by setting bCacheNews to true

*////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////


#property copyright "Paul Hampton-Smith"
#property link      ""

#include <stdlib.mqh>
#include <stderror.mqh>

string strNewsFileName = "NewsItems.csv"; // located in experts\files
int chNewsFileSeparator = ','; // to work in .csv format

#define MAX_NEWS_ITEMS  50
#define MAX_PARAMETERS  5

bool bNewsTrackerDebug = false; // switch on for lots more Print messages
static bool bNewsIsCached = false;
extern bool bCacheNews = false;


//////////////////////////////////////////////
// MAIN FUNCTIONS
//////////////////////////////////////////////


bool NewsEvent(int nWithinMinutes, string& strDescription, datetime& NewsTimeServer)
{
	// returns true if news event is nWithinMinutes
	// also returns description and news time
	
	double dblParameters[MAX_PARAMETERS];
   NextNews(strDescription, NewsTimeServer, dblParameters, nWithinMinutes);

   bool bReturn = CurTime() >= NewsTimeServer - nWithinMinutes*60 && 
               CurTime() <= NewsTimeServer + nWithinMinutes*60;
   if (bNewsTrackerDebug) Print("NewsEvent(",nWithinMinutes,") returning ",bToStr(bReturn)," and (\"",strDescription,"\",",TimeToStr(NewsTimeServer),")");
   return(bReturn);
}

bool NextNews(string& strDescription, datetime& EarliestNextTime, double& dblParameters[], int nMinutesInThePast = 0)
{
	// returns next news item details. If none, returns false

   static bool bDontBugMe = false;
   bool bReturn;
	int i;
	   
   static string strDescriptionArray[MAX_NEWS_ITEMS];
   static string strCurrencyArray[MAX_NEWS_ITEMS];
   static datetime dtServerArray[MAX_NEWS_ITEMS];
   static double dblParameters2D[MAX_NEWS_ITEMS][MAX_PARAMETERS];
   static int nNewsItemCount = 0;
   
   string strCurrency;
   datetime VeryFutureTime = StrToTime("2030.01.01 00:00");
   EarliestNextTime = VeryFutureTime;
   datetime DateTimeServer;

   if (!bNewsIsCached || !bCacheNews)
   {
   	nNewsItemCount = 0;
      int handle = FileOpen(strNewsFileName, FILE_CSV|FILE_READ, chNewsFileSeparator);
      if (handle <= 0)
      {   
         // only Alert the user once to the fact that strNewsFileName is missing
         if (!bDontBugMe)
         {
            if (IsTesting()) Alert("Please add file ",strNewsFileName," to tester\files folder, see AvoidTheNews.mqh for format");
            else Alert("Please add file ",strNewsFileName," to experts\files folder, see AvoidTheNews.mqh for format");
            bDontBugMe = true;
         }
         // repeated entries into log
         if (IsTesting()) Print("Please add file ",strNewsFileName," to tester\files folder, see AvoidTheNews.mqh for format");
         else Print("Please add file ",strNewsFileName," to experts\files folder, see AvoidTheNews.mqh for format");
         return(0);
      }
   
      // read header line
      while(!FileIsLineEnding(handle)) FileReadString(handle);
 
      // read news date lines and cache in arrays
      string strDescr, strCurr; datetime dtServer; double dblPar[MAX_PARAMETERS];
      bool bReadOK = ReadCSVline(handle, strDescr, strCurr, dtServer, dblPar);
      while (bReadOK)
      {
         strDescriptionArray[nNewsItemCount] = StringConcatenate(strCurr," - ",strDescr);
         strCurrencyArray[nNewsItemCount] = strCurr;
         dtServerArray[nNewsItemCount] = dtServer;
         for (i = 0 ; i < MAX_PARAMETERS ; i++)
         {
         	dblParameters2D[nNewsItemCount][i] = dblPar[i];
         }
			nNewsItemCount++;         
			bReadOK = ReadCSVline(handle, strDescr, strCurr, dtServer, dblPar);
      }
      FileClose(handle);
      if (bNewsTrackerDebug) Print("News is cached with ",nNewsItemCount," items");
		bNewsIsCached = true;     
   }

	// search for next news item for the currency pair
	int nEarliestNextTimeIndex = -1;
   for (i = 0 ; i < nNewsItemCount ; i++)
   {
      strCurrency = strCurrencyArray[i]; DateTimeServer = dtServerArray[i];
      if (CurTime()-nMinutesInThePast*60 > DateTimeServer || !CurrencyPairContains(strCurrency) ) continue;
      if (EarliestNextTime > DateTimeServer)
      {
         EarliestNextTime = DateTimeServer;
         nEarliestNextTimeIndex = i;
        	bReturn = true;
      }
   }

   if (nEarliestNextTimeIndex == -1)  
   {
      // No dates in the future
      EarliestNextTime = 0;
      DateTimeServer = 0;
      strDescription = "";
      bReturn = false;
   }
   else
   {
   	// found a future news item. Copy details to be returned
      strDescription = strDescriptionArray[nEarliestNextTimeIndex];
	   for (i = 0 ; i < MAX_PARAMETERS ; i++)
      {
      	dblParameters[i] = dblParameters2D[nEarliestNextTimeIndex][i];
      }
   }

   if (bNewsTrackerDebug) Print("Function NextNews(",nMinutesInThePast,") returning ",bToStr(bReturn)," and (",TimeToStr(EarliestNextTime),",\"",strDescription,"\",",ArrayTrace(dblParameters),")");
   return(bReturn);
}  





//////////////////////////////////////////////
// SUPPORTING FUNCTIONS
//////////////////////////////////////////////

bool ReadCSVline(int handle, string& strDescription, string& strCurrency, datetime& DateTimeServer, double& dblParameters[])
{
   int nLastError = 0;

//////////////////////////////////////
   // Handle

   if (handle <= 0)
   {
      Print("Invalid file handle for news file", strNewsFileName);
      if (bNewsTrackerDebug) Print("Function ReadCSVline() returning false");
      return(false);
   }

//////////////////////////////////////
   // Description
   strDescription = FileReadString(handle); nLastError = GetLastError();
   if (nLastError != 0 && nLastError != ERR_NOT_INITIALIZED_STRING)
   {
      if (nLastError != ERR_END_OF_FILE) Print(ErrorDescription(nLastError)," reading description from ",strNewsFileName);
      if (bNewsTrackerDebug) Print("Function ReadCSVline() returning false");
      return(false);
   }

//////////////////////////////////////
   // Currency
   strCurrency = FileReadString(handle); nLastError = GetLastError();
   if (nLastError != 0 && nLastError != ERR_NOT_INITIALIZED_STRING)
   {
      Print(ErrorDescription(nLastError)," reading currency from ",strNewsFileName);
      if (bNewsTrackerDebug) Print("Function ReadCSVline() returning false");
      return(false);
   }
   if (StringLen(strCurrency) != 3)
   {
      Print("Problem with strCurrency = \"",strCurrency,"\" in ",strNewsFileName,". Should be 3 characters");
      if (bNewsTrackerDebug) Print("Function ReadCSVline() returning false");
      return(false);
   }

//////////////////////////////////////
   // DateTime
   string strDateTimeFromFile = FileReadString(handle); nLastError = GetLastError();
   if (nLastError != 0 && nLastError != ERR_NOT_INITIALIZED_STRING)
   {
      Print(ErrorDescription(nLastError)," reading DateTime from ",strNewsFileName);
      if (bNewsTrackerDebug) Print("Function ReadCSVline() returning false");
      return(false);
   }
   
//////////////////////////////////////
   // Timezone
   string strTimeZone = FileReadString(handle); nLastError = GetLastError();
   if (nLastError != 0 && nLastError != ERR_NOT_INITIALIZED_STRING)
   {
      Print(ErrorDescription(nLastError)," reading TimeZone from ",strNewsFileName);
      if (bNewsTrackerDebug) Print("Function ReadCSVline() returning false");
      return(false);
   }

//////////////////////////////////////
   // Convert DateTime
   datetime DateTimeFromFile = MyStrToTime(strDateTimeFromFile);
   if (DateTimeFromFile == 0)
   {
      Print("Problem with DateTime = \"",strDateTimeFromFile,"\" from ",strNewsFileName);
      if (bNewsTrackerDebug) Print("Function ReadCSVline() returning false");
      return(false);
   }
   DateTimeServer = UTCtoServer(TimeToUTC(DateTimeFromFile, strTimeZone));
   
//////////////////////////////////////
   // Parameters
   int nParameterPos = 0;
   while(!FileIsLineEnding(handle)) 
   {
   	dblParameters[nParameterPos] = FileReadNumber(handle);
	   if (nLastError != 0)
   	{
      	Print(ErrorDescription(nLastError)," reading dblParameters[",nParameterPos,"] from ",strNewsFileName);
      	if (bNewsTrackerDebug) Print("Function ReadCSVline() returning false");
      	return(false);
   	}
   	nParameterPos++;
  	}

   if (bNewsTrackerDebug) Print("Function ReadCSVline() returning true and (\"",strDescription,"\",\"",strCurrency,"\",",TimeToStr(DateTimeServer),",",ArrayTrace(dblParameters),")");
   return(true);
}



string CleanFilename(string str)
{
	// change illegal filename characters in str to '-'
	int nPos; 

	string strRet = str;
	nPos = StringFind(str,"/");
	if (nPos != -1) strRet = StringSetChar(strRet, nPos, '-');
	
	nPos = StringFind(strRet,"\\");
	if (nPos != -1) strRet = StringSetChar(strRet, nPos, '-');
	
	nPos = StringFind(strRet,":");
	if (nPos != -1) strRet = StringSetChar(strRet, nPos, '-');

	nPos = StringFind(strRet,"*");
	if (nPos != -1) strRet = StringSetChar(strRet, nPos, '-');

	nPos = StringFind(strRet,"?");
	if (nPos != -1) strRet = StringSetChar(strRet, nPos, ' ');

	nPos = StringFind(strRet,"\"");
	if (nPos != -1) strRet = StringSetChar(strRet, nPos, '-');

	nPos = StringFind(strRet,">");
	if (nPos != -1) strRet = StringSetChar(strRet, nPos, '-');

	nPos = StringFind(strRet,"<");
	if (nPos != -1) strRet = StringSetChar(strRet, nPos, '-');

	nPos = StringFind(strRet,"|");
	if (nPos != -1) strRet= StringSetChar(strRet, nPos, '-');

	return(strRet);
}

string ArrayTrace(double arr[])
{
	// returns "(arr[0],arr[1],...)"
	string strRet = StringConcatenate(arr[0],"");
  	for (int i = 1 ; i < ArraySize(arr) ; i++)
  	{
   	strRet = StringConcatenate(strRet,",",arr[i]);
   }
   return(strRet);
}

string bToStr(bool b)
{
   if (b) return("true"); 
   else return("false");
}

datetime MyStrToTime(string strTime)
{
   // MT4 StrToTime doesn't adds actual testing date to time when using "hh:mm" variation
 
   if (!IsTesting() || StringLen(strTime) > 5) 
   {
      // works fine when not testing and with yyyy.mm.dd hh:mm format
      return(StrToTime(strTime));
   }
   else
   {
      // use actual CurTime() date
      return(StrToTime(TimeToStr(CurTime(),TIME_DATE) + " " + strTime));
   }
}

string MyTimeToString(datetime dt)
{
   // output format dd/mm/yyyy mm:hh, eg for reading by Excel
   string strDayPad;
   int nDay = TimeDay(dt);
   if (nDay >= 0 && nDay <= 9) strDayPad = "0";
   else strDayPad = "";

   string strMonthPad;
   int nMonth = TimeMonth(dt);
   if (nMonth >= 0 && nMonth <= 9) strMonthPad = "0";
   else strMonthPad = "";
   
   string strHourPad;
   int nHour = TimeHour(dt);
   if (nHour >= 0 && nHour <= 9) strHourPad = "0";
   else strHourPad = "";
   
   string strMinutePad;
   int nMinute = TimeMinute(dt);
   if (nMinute >= 0 && nMinute <= 9) strMinutePad = "0";
   else strMinutePad = "";

   string strSecondPad;
   int nSecond = TimeSeconds(dt);
   if (nSecond >= 0 && nSecond <= 9) strSecondPad = "0";
   else strSecondPad = "";

   return(StringConcatenate(strDayPad,nDay,"/",strMonthPad,nMonth,"/",TimeYear(dt)," ",strHourPad,nHour,":",strMinutePad,nMinute,":",strSecondPad,nSecond));
}

datetime UTCtoServer(datetime UTC)
{
   // Uses global variable UTCtoServerTimeInMinutes to convert UTC to server time.
   // this enables different instances of MT4 using different brokers to run EAs with identical UTC time triggers
   // Examples:
   //    UTCtoServerTimeInMinutes = 0   for InterbankFX because their server time = UTC
   //    UTCtoServerTimeInMinutes = 180 for FXDD        because their server time = UTC plus 3 hours

   static bool bDontBugMe = false;
   if (!GlobalVariableCheck("UTCtoServerTimeInMinutes"))
   {
      // only Alert the user once to the fact that this variable is missing
      if (!bDontBugMe)
      {
         Alert("Please add global variable UTCtoServerTimeInMinutes (menu item Tools->Global Variables)");
         bDontBugMe = true;
      }
      // repeated entries into log
      Print("Please add global variable UTCtoServerTimeInMinutes (menu item Tools->Global Variables)");
      return(0);
   }
   if (bNewsTrackerDebug) Print("UTCtoServer(",TimeToStr(UTC),") returning ",TimeToStr(UTC+60*GlobalVariableGet("UTCtoServerTimeInMinutes")));
   return(UTC+60*GlobalVariableGet("UTCtoServerTimeInMinutes"));
}

datetime ServerToUTC(datetime dtServer)
{
   // Uses global variable UTCtoServerTimeInMinutes to convert server time to UTC.

   static bool bDontBugMe = false;
   if (!GlobalVariableCheck("UTCtoServerTimeInMinutes"))
   {
      // only Alert the user once to the fact that this variable is missing
      if (!bDontBugMe)
      {
         Alert("Please add global variable UTCtoServerTimeInMinutes (menu item Tools->Global Variables)");
         bDontBugMe = true;
      }
      // repeated entries into log
      Print("Please add global variable UTCtoServerTimeInMinutes (menu item Tools->Global Variables)");
      return(0);
   }
   if (bNewsTrackerDebug) Print("ServerToUTC(",TimeToStr(dtServer),") returning ",TimeToStr(dtServer-60*GlobalVariableGet("UTCtoServerTimeInMinutes")));
   return(dtServer-60*GlobalVariableGet("UTCtoServerTimeInMinutes"));
}

datetime TimeToUTC(datetime dt, string strTimeZone)
{
   int BST_TimeZone = 1;
   int EST_TimeZone = -4;
   int EDT_TimeZone = -5;
   double CST_TimeZone = 9.5;
   double CSST_TimeZone = 10.5;

   datetime dtReturn;
   if      (strTimeZone == "UTC") dtReturn = dt;
   else if (strTimeZone == "GMT") dtReturn = dt;
   else if (strTimeZone == "BST") dtReturn = dt - BST_TimeZone*3600;
   else if (strTimeZone == "EST") dtReturn = dt - EST_TimeZone*3600;
   else if (strTimeZone == "EDT") dtReturn = dt - EDT_TimeZone*3600;
   else if (strTimeZone == "CST") dtReturn = dt - CST_TimeZone*3600;
   else if (strTimeZone == "CSST") dtReturn = dt - CSST_TimeZone*3600;
   else 
   { 
      Print("Unknown timezone ",strTimeZone, " in TimeToUTC()");
      dtReturn = 0;
   }
   
   if (bNewsTrackerDebug) Print("TimeToUTC(",TimeToStr(dt),",",strTimeZone,") returning ",TimeToStr(dtReturn));
   return(dtReturn);
}      

bool CurrencyPairContains(string strCurrency)
{
   bool bReturn = StringSubstr(Symbol(),0,3) == strCurrency || StringSubstr(Symbol(),3,3) == strCurrency;
   if (bNewsTrackerDebug) Print("Function CurrencyPairContains(",strCurrency,") returning ",bToStr(bReturn));
   return(bReturn);
}

datetime ServerToLocal(datetime dtServer)
{
   int nRoundedMinutes = (LocalTime()-CurTime()+30)/60;
   return(dtServer + nRoundedMinutes*60);
}

